import os
import traceback
import math
import argparse
import numpy as np
import time
import functools
from typing import Tuple, List, Dict, Any, Set
import logging

import xml.etree.ElementTree as xmlet
import numpy as np
import csv

import re

import time
from datetime import date
#import openpyxl
#from openpyxl.styles import Border, Side, Font, Alignment

from PyQt5.QtCore import pyqtSignal, QPointF, Qt, QCoreApplication
from PyQt5 import QtGui
import PyQt5.QtWidgets as qt
from PyQt5 import QtCore

import UsefullVariables as vrb
import UsefullWidgets as wgt
import UsefullFunctions as fct
import DatabaseFunction as Dfct
import UsefullDisplay as Dsp

import PyIPSDK
import PyIPSDK.IPSDKIPLAdvancedMorphology as advmorpho
import PyIPSDK.IPSDKIPLUtility as util
import PyIPSDK.IPSDKIPLBinarization as bin
import PyIPSDK.IPSDKIPLArithmetic as arithm
import PyIPSDK.IPSDKIPLLogical as logic
import PyIPSDK.IPSDKIPLGlobalMeasure as glbmsr

from UsefullWidgets import Measure

import PyIPSDK.IPSDKUI as ui

import sys
import pandas as pd


mainWindow = vrb.mainWindow

try:
    groupMenu = mainWindow.groupMenu

    buttonProcessMTortuosity = wgt.PushButtonImage(vrb.folderMacroInterface + "/Macro_MTortuosity/mtortuosity.png")
    buttonProcessMTortuosity.setFixedSize(30 * vrb.ratio, 30 * vrb.ratio)
    buttonProcessMTortuosity.setToolTip("Compute M-Tortuosity")

    groupMenu.layoutBar1.addWidget(buttonProcessMTortuosity, 0, vrb.numMacro, Qt.AlignVCenter)
    vrb.numMacro += 1
except:
    pass


class MTortuosityWidget(qt.QWidget):

    def __init__(self):
        qt.QWidget.__init__(self)

        try:
            self.setWindowFlag(Qt.WindowStaysOnTopHint)
        except:
            self.setWindowFlags(Qt.WindowStaysOnTopHint)

        self.stopProcess = False

        self.curImg=None
        self.date = date.today()

        # MODE TYPE
        self.labelModeType = qt.QLabel("Mode :")
        self.labelModeType.setFixedWidth(70 * vrb.ratio)
        self.comboBoxModeType = qt.QComboBox(self)
        self.comboBoxModeType.setEditable(True)
        self.comboBoxModeType.setFixedSize(90 * vrb.ratio, 20 * vrb.ratio)
        self.comboBoxModeType.currentIndexChanged.connect(self.changeMode)

        # NB SEEDS (for random mode)
        self.labelNbSeeds = qt.QLabel("Number random seeds :")
        self.labelNbSeeds.setFixedWidth(120 * vrb.ratio)
        self.lineEditNbSeeds = qt.QLineEdit()
        self.lineEditNbSeeds.setText("3")
        self.lineEditNbSeeds.setFixedWidth(50 * vrb.ratio)
        self.lineEditNbSeeds.setAlignment(Qt.AlignRight)

        # SEEDS REPRODUCTIBILITY
        self.labelReproductibility = qt.QLabel("Seeds for reproductibility :")
        self.labelReproductibility.setFixedWidth(120 * vrb.ratio)
        self.lineEditReproductibility = qt.QLineEdit()
        self.lineEditReproductibility.setText("32")
        self.lineEditReproductibility.setAlignment(Qt.AlignRight)
        self.lineEditReproductibility.setFixedWidth(50 * vrb.ratio)

        # BACKUP DIRECTORY
        self.fileSelectorSave = FileSelectorSave()
        self.fileSelectorSave.setFixedHeight(50 * vrb.ratio)

        # CSV DIRECTORY
        self.fileSelectorCsv = FileSelectorCSV()
        self.fileSelectorCsv.setFixedHeight(50 * vrb.ratio)
        self.fileSelectorCsv.setVisible(False)

        # SAVE IMAGES CHECK BOX
        self.radioButtonSaveImg = qt.QRadioButton("Save images result")
        self.radioButtonSaveImg.setFixedSize(150 * vrb.ratio, 20 * vrb.ratio)
        self.radioButtonSaveImg.setChecked(False)

        # PARAMS GROUPBOX
        self.groupBoxParams = qt.QGroupBox()
        layoutParams = qt.QGridLayout()
        layoutParams.setContentsMargins(10,10,10,10)
        layoutParams.addWidget(self.labelModeType, 0, 0, Qt.AlignLeft)
        layoutParams.addWidget(self.comboBoxModeType, 0, 1, Qt.AlignRight)
        layoutParams.addWidget(self.labelNbSeeds, 1, 0, Qt.AlignLeft)
        layoutParams.addWidget(self.lineEditNbSeeds, 1, 1, Qt.AlignRight)
        layoutParams.addWidget(self.labelReproductibility, 2, 0, Qt.AlignLeft)
        layoutParams.addWidget(self.lineEditReproductibility, 2, 1, Qt.AlignRight)
        self.groupBoxParams.setLayout(layoutParams)
        self.groupBoxParams.setFixedSize(260*vrb.ratio, 150*vrb.ratio)
        self.groupBoxParams.setTitle("Parameters :")
        self.groupBoxParams.setStyleSheet('QGroupBox:title {left: 20px ;padding-left: 10px;padding-right: 10px; padding-top: -12px; color:rgb(6, 115, 186)}  QGroupBox {font: bold; border: 1px solid gray; margin-top: 12 px}')

        # FILES GROUPBOX
        self.groupBoxFiles = qt.QGroupBox()
        layoutFiles = qt.QGridLayout()
        layoutFiles.setContentsMargins(10, 10, 10, 10)
        layoutFiles.addWidget(self.fileSelectorSave, 0, 0)
        layoutFiles.addWidget(self.fileSelectorCsv, 1, 0)
        layoutFiles.addWidget(self.radioButtonSaveImg, 2, 0)
        self.groupBoxFiles.setLayout(layoutFiles)
        self.groupBoxFiles.setFixedSize(260 * vrb.ratio, 150 * vrb.ratio)

        # OK BUTTON
        self.buttonOK = wgt.PushButtonImage(vrb.folderImages + "/Validate.png")
        self.buttonOK.setFixedSize(25 * vrb.ratio, 25 * vrb.ratio)

        # HELP BUTTON
        self.buttonHelp = wgt.PushButtonImage(vrb.folderImages + "/Interrogation.png")
        self.buttonHelp.setStyleSheet("background-color : transparent; border :0px")
        self.buttonHelp.setFixedSize(25 * vrb.ratio, 25 * vrb.ratio)

        # BUTTONS GROUPBOX
        groupBoxValidateHelpButtons = qt.QGroupBox()
        layoutValidateHelpButtons = qt.QGridLayout()
        layoutValidateHelpButtons.setContentsMargins(0, 0, 0, 0)
        layoutValidateHelpButtons.addWidget(self.buttonHelp, 0, 0, Qt.AlignLeft)
        layoutValidateHelpButtons.addWidget(self.buttonOK, 0, 1, Qt.AlignRight)
        groupBoxValidateHelpButtons.setLayout(layoutValidateHelpButtons)
        groupBoxValidateHelpButtons.setFixedWidth(260 * vrb.ratio)

        self.layout = qt.QGridLayout()
        self.layout.addWidget(self.groupBoxParams, 0,0)
        self.layout.addWidget(self.groupBoxFiles, 1,0)
        self.layout.addWidget(groupBoxValidateHelpButtons, 2, 0)

        self.setLayout(self.layout)

        self.fileSelectorSave.buttonOpen.clicked.connect(self.openQDialogSave)
        self.fileSelectorCsv.buttonOpen.clicked.connect(self.openQDialogCsv)
        self.buttonOK.clicked.connect(self.m_tortuosity_binary)
        self.buttonHelp.clicked.connect(self.openReadMe)


        try:
            vrb.mainWindow.groupBoxProcessing.buttonStop.clicked.connect(self.stopProcessFunction)
        except:
            pass

        style = fct.getStyleSheet()
        self.setStyleSheet(style)

        self.setWindowTitle("M-Tortuosity Module")
        self.setFixedSize(280 * vrb.ratio, 350 * vrb.ratio)

    # opens the read me file in the macro folder
    def openReadMe(self):
        osCommandString = "notepad.exe " + str(vrb.folderMacroInterface) + "/Macro_MTortuosity/README.txt"
        os.system(osCommandString)

    # changes the display of the window depending on the mode selected
    def changeMode(self):
        if (self.comboBoxModeType.currentIndex() == 0):
            self.fileSelectorCsv.setVisible(False)
            self.labelNbSeeds.setVisible(True)
            self.lineEditNbSeeds.setVisible(True)
            self.groupBoxParams.setFixedSize(260 * vrb.ratio, 150 * vrb.ratio)
            self.groupBoxFiles.setFixedSize(260 * vrb.ratio, 150 * vrb.ratio)
        else:
            self.fileSelectorCsv.setVisible(True)
            self.labelNbSeeds.setVisible(False)
            self.lineEditNbSeeds.setVisible(False)
            self.groupBoxParams.setFixedSize(260 * vrb.ratio, 110 * vrb.ratio)
            self.groupBoxFiles.setFixedSize(260 * vrb.ratio, 170 * vrb.ratio)
        self.update()

    def stopProcessFunction(self):
        self.close()
        #self.stopProcess = True

    def loadModeType(self):
        self.comboBoxModeType.clear()
        self.comboBoxModeType.addItem("Random")
        self.comboBoxModeType.addItem("Deterministic")
        self.comboBoxModeType.setCurrentIndex(0)

    def openQDialogSave(self):
        defaultFolder = Dfct.childText(vrb.userPathElement, "CustomImport")
        if defaultFolder is None or defaultFolder == "" or os.path.exists(defaultFolder) == False:
            defaultFolder = "C:/"
        filename = qt.QFileDialog.getExistingDirectory(self, "Select backup folder", defaultFolder)
        if filename != "" and filename != None:
            try:
                self.fileSelectorSave.lineEditFile.setText(filename)
                Dfct.SubElement(vrb.userPathElement, "CustomImportSave").text = os.path.dirname(filename)
                Dfct.saveXmlElement(vrb.userPathElement, vrb.folderInformation + "/UserPath.mho", forceSave=True)
            except:
                pass

    def openQDialogCsv(self):
        defaultFolder = Dfct.childText(vrb.userPathElement, "CustomImport")
        if defaultFolder is None or defaultFolder == "" or os.path.exists(defaultFolder) == False:
            defaultFolder = "C:/"
        #filename = qt.QFileDialog.getExistingDirectory(self, "Sélectionnez le fichier CSV", defaultFolder)
        filename = qt.QFileDialog.getOpenFileName(self, "Open CSV Data File", "", "CSV data files (*.csv)")
        if filename[0] != "" and filename[0] != None:
            try:
                self.fileSelectorCsv.lineEditFile.setText(filename[0])
                # Dfct.SubElement(vrb.userPathElement, "CustomImportCsv").text = os.path.dirname(filename[0])
                # Dfct.saveXmlElement(vrb.userPathElement, vrb.folderInformation + "/UserPath.mho", forceSave=True)
            except:
                pass

    def sampling(self, shape: Tuple[int], n_points: int) -> Set[Tuple[Any, ...]]:
        points = set(
            tuple(map(lambda size: int(np.random.uniform(0, size)), shape))
            for _ in range(n_points)
        )
        return points

    def euclidean_dist(self, p1: Tuple[Any, ...], p2: Tuple[Any, ...]) -> float:
        return math.sqrt(sum(map(lambda elt: pow(elt[0] - elt[1], 2), zip(p1, p2))))

    def center_of_mass(self, in_binary: Any) -> Tuple[int]:
        outFormulaX = arithm.formula3dImg("x*I1", InOptImg3d1=in_binary)
        outFormulaY = arithm.formula3dImg("y*I1", InOptImg3d1=in_binary)
        outFormulaZ = arithm.formula3dImg("z*I1", InOptImg3d1=in_binary)
        statFormulaX = glbmsr.statsMsr3d(outFormulaX)
        statFormulaY = glbmsr.statsMsr3d(outFormulaY)
        statFormulaZ = glbmsr.statsMsr3d(outFormulaZ)
        statBinary = glbmsr.statsMsr3d(in_binary)
        x = statFormulaX.sum / statBinary.sum
        y = statFormulaY.sum / statBinary.sum
        z = statFormulaZ.sum / statBinary.sum
        return x, y, z

    def parse_csv_points_file(self, csv_points_file: str) -> Set[Tuple[Any, ...]]:
        points_df = pd.read_csv(csv_points_file)
        assert "z" in points_df.columns
        assert "y" in points_df.columns
        assert "x" in points_df.columns
        return set(map(lambda point: tuple(point), points_df.values.tolist()))

    def get_seed_points(self, shape: Tuple[int]) -> Set[Tuple[Any, ...]]:
        #if mode == deterministic
        if self.comboBoxModeType.currentIndex() == 1:
            deterministic_points = self.parse_csv_points_file(self.fileSelectorCsv.lineEditFile.text())

            def _check_coordinates(point):
                return all(p < s and p >= 0 for p, s in zip(point, shape))

            filtered_set = set(
                filter(lambda p: _check_coordinates(p), deterministic_points)
            )
            if len(deterministic_points - filtered_set):
                print(
                    f"Points {(deterministic_points - filtered_set)}  discarded because resides out of image bounds."
                    f"Image shape is {shape}"
                )
            return filtered_set
        elif self.comboBoxModeType.currentIndex() == 0:
            return self.sampling(shape, int(self.lineEditNbSeeds.text()))
        else:
            raise ValueError("Mode should be either deterministic or random")

    # computes the tortuosity of a binary image
    def m_tortuosity_binary(self):
        """ "
        Compute m_tortuosity on a binary image
        """
        try:
            vrb.mainWindow.toggleGroupBoxProcessing()
            vrb.mainWindow.groupBoxProcessing.setText('')
            vrb.mainWindow.groupBoxProcessing.buttonStop.setVisible(True)
            vrb.mainWindow.groupBoxProcessing.progressBar.setVisible(False)
            vrb.mainWindow.groupBoxProcessing.labelLeft.setFixedWidth(190 * vrb.ratio)
        except:
            pass

        self.errorOccured=False
        saveExcel = True

        try:
            nbSeeds = int(self.lineEditNbSeeds.text())
            reproductibilitySeeds = int(self.lineEditReproductibility.text())
            if (nbSeeds < 3 or nbSeeds > 20):
                self.messageBox = wgt.MessageBox(
                    "The number of seeds points must be defined between 3 and 20 seeds.", '',
                    buttons=[qt.QMessageBox.Ok], icon=qt.QMessageBox.Warning, windowTitle="Error")
                self.messageBox.exec()
                self.errorOccured = True

            elif (reproductibilitySeeds < 32 or reproductibilitySeeds > 128):
                self.messageBox = wgt.MessageBox(
                    "The number of seeds for reproductibility must be defined between 32 and 128 seeds.", '',
                    buttons=[qt.QMessageBox.Ok], icon=qt.QMessageBox.Warning, windowTitle="Error")
                self.messageBox.exec()
                self.errorOccured = True
        except Exception:
            self.messageBox = wgt.MessageBox(
                "Seeds number must be integers.", '',
                buttons=[qt.QMessageBox.Ok], icon=qt.QMessageBox.Warning, windowTitle="Error")
            self.messageBox.exec()
            self.errorOccured = True
            pass

        if (self.fileSelectorSave.lineEditFile.text() == ""):
            if self.radioButtonSaveImg.isChecked():
                self.messageBox = wgt.MessageBox(
                    "You must select a backup directory.", '',
                    buttons=[qt.QMessageBox.Ok], icon=qt.QMessageBox.Warning, windowTitle="Error")
                self.messageBox.exec()
                self.errorOccured = True
            else:
                saveExcel = False

        elif (self.comboBoxModeType.currentIndex()==1 and self.fileSelectorCsv.lineEditFile.text() == ""):
            self.messageBox = wgt.MessageBox(
                "You must select a CSV file with deterministic mode.", '',
                buttons=[qt.QMessageBox.Ok], icon=qt.QMessageBox.Warning, windowTitle="Error")
            self.messageBox.exec()
            self.errorOccured = True

        if not self.errorOccured:

            try:
                self.close()
                start = time.time()
                firstRow = 13
                lines = []

                resultFolder = self.fileSelectorSave.lineEditFile.text()

                # For reproductibility
                np.random.seed(int(self.lineEditReproductibility.text()))

                try:
                    self.curImg = vrb.mainWindow.currentLabel.image
                    self.curImgName = vrb.mainWindow.currentLabel.name
                except:
                    self.messageBox = wgt.MessageBox(
                        "No image to process.", '',
                        buttons=[qt.QMessageBox.Ok], icon=qt.QMessageBox.Warning, windowTitle="Error")
                    self.messageBox.exec()
                    self.close()


                # Taille de l'image courante
                sizeX = self.curImg.getSizeX()
                sizeY = self.curImg.getSizeY()
                if (self.curImg.getVolumeGeometryType() == PyIPSDK.eVGT_3d):
                    sizeZ = self.curImg.getSizeZ()

                if saveExcel and not os.path.exists(resultFolder):
                    self.messageBox = wgt.MessageBox(
                        "Backup directory is not valid : " + self.fileSelectorSave.lineEditFile.text(), '',
                        buttons=[qt.QMessageBox.Ok], icon=qt.QMessageBox.Warning, windowTitle="Error")
                    self.messageBox.exec()
                else:

                    user = os.path.expanduser('~')

                    # initialization of attributes processing
                    #savePath = resultFolder + "/MTortuosity_results.xlsx"
                    savePath = resultFolder + "/MTortuosity_results.txt"
                    # try:
                    #     # load the excel file
                    #     workbook = openpyxl.load_workbook(
                    #         user + '/AppData/Roaming/ReactivIP/Explorer/Macro_Interface/Macro_MTortuosity/MTortuosity_results.xlsx')
                    # except:
                    #     traceback.print_exc(file=sys.stderr)
                    #     print("WARNING : Unable to import reference excel model.")
                    #     self.messageBox = wgt.MessageBox(
                    #         "WARNING : Unable to import reference excel model.", '',
                    #         buttons=[qt.QMessageBox.Ok], icon=qt.QMessageBox.Warning, windowTitle="Error")
                    #     self.messageBox.exec()

                    # sheet_results = workbook["Results"]
                    #
                    # sheet_results['B3'] = str(self.date.strftime("%d/%m/%y"))
                    # sheet_results['B4'] = str(self.curImgName)
                    # if (self.curImg.getVolumeGeometryType() == PyIPSDK.eVGT_3d):
                    #     sheet_results['B5'] = str(sizeX) + "x" + str(sizeY) + "x" + str(sizeZ)
                    # else:
                    #     sheet_results['B5'] = str(sizeX) + "x" + str(sizeY)
                    #
                    # if(self.comboBoxModeType.currentIndex()==0):
                    #     sheet_results['B6'] = "None"
                    # else:
                    #     sheet_results['B6'] = self.fileSelectorCsv.lineEditFile.text()
                    #
                    # sheet_results['C8'] = self.comboBoxModeType.currentText()
                    # sheet_results['C10'] = self.lineEditReproductibility.text()

                    vrb.mainWindow.groupBoxProcessing.setText("Processing ...")
                    qt.QApplication.processEvents()

                    # Load input (binary) image
                    in_binary = self.curImg

                    # Convert to appropriate IPSDK data type
                    in_binary = bin.lightThresholdImg(in_binary, 1.0)

                    # Random points sampling
                    points = self.get_seed_points(in_binary.array.shape)

                    # Allocate working array
                    working_array = np.zeros(in_binary.array.shape, dtype=bool)
                    seed_binary = PyIPSDK.fromBinaryArray(working_array)

                    # Find valid points set
                    valid_points = set(filter(lambda p: in_binary.array[p] != 0, points))

                    # if (self.comboBoxModeType.currentIndex() == 0):
                    #     sheet_results['C9'] = self.lineEditNbSeeds.text()
                    # else:
                    #     sheet_results['C9'] = len(valid_points)

                    if len(points - valid_points):
                        print(
                            f"Points {points - valid_points}  discarded because resides in background."
                        )
                        lines.append(f"Points {points - valid_points}  discarded because resides in background.\n")
                    print(f"Valid seed Points list for computation :  {valid_points}")
                    lines.append(f"Valid seed Points list for computation :  {valid_points}\n")
                    vrb.mainWindow.groupBoxProcessing.setText(f"Points {points - valid_points}  discarded because resides in background.")
                    qt.QApplication.processEvents()
                    vrb.mainWindow.groupBoxProcessing.setText(f"Valid seed Points list for computation :  {valid_points}")
                    qt.QApplication.processEvents()
                    """if len(valid_points) == 1:
                        logger.error(
                            f"Only one valid point after filtering, please change your inputs."
                        )
                        raise Exception(
                            f"Only one valid point after filtering, please change your inputs."
                        )"""

                    # Weight for IPSDK genericSeededDistanceMap3dImg ( scale x10)
                    scale_factor = 10
                    inDistWeight3d = PyIPSDK.createDistWeight3d(
                        scale_factor,
                        scale_factor,
                        scale_factor,
                        scale_factor * math.sqrt(2),
                        scale_factor * math.sqrt(2),
                        scale_factor * math.sqrt(2),
                        scale_factor * math.sqrt(3),
                    )

                    c_n = []
                    nbDistMap=0
                    table = []
                    table.append(["Seed number","Seed coordinates","M-coefficient",""])
                    outDistImg = None
                    for p in valid_points:
                        nbDistMap+=1
                        print(f"M-coefficient computation for point {p}")
                        lines.append(f"M-coefficient computation for point {p}\n")
                        vrb.mainWindow.groupBoxProcessing.setText(
                            f"M-coefficient computation for point {nbDistMap} / {len(valid_points)}")
                        qt.QApplication.processEvents()
                        # Init seed
                        seed_binary.array[p] = 1

                        startDistMap = time.time()
                        # Seeded distance map computation
                        if outDistImg is None:
                            outDistImg = advmorpho.genericSeededDistanceMap3dImg(
                                in_binary, seed_binary, inDistWeight3d
                            )
                        else:
                            advmorpho.genericSeededDistanceMap3dImg(
                                in_binary, seed_binary, inDistWeight3d, outDistImg
                            )

                        endDistMap = time.time()
                        print("Distance map : " + str(endDistMap-startDistMap))
                        lines.append("Elapsed time for distance map : " + str(endDistMap-startDistMap) + " seconds.\n")

                        if self.radioButtonSaveImg.isChecked():
                            # SAVE TIFF IMAGES
                            startSaveImg = time.time()
                            PyIPSDK.saveTiffImageFile(self.fileSelectorSave.lineEditFile.text()+"/outDistImg_seed"+ str(p) + ".tif", outDistImg)
                            endSaveImg = time.time()
                            print("Save Image : " + str(endSaveImg-startSaveImg))
                            lines.append("Time elapsed for saving image result outDistImg_seed" + str(p) +" : " + str(endSaveImg-startSaveImg) + " seconds.\n")

                        # downscale distances map result by 10
                        downscaleImg = arithm.multiplyScalarImg(outDistImg, 1 / scale_factor)

                        # Tortuosity computation from p for all pp
                        pp_tortuosity = [
                            (
                                pp,
                                downscaleImg.array[pp] / self.euclidean_dist(p, pp),
                            )
                            for pp in valid_points
                            if (pp != p and downscaleImg.array[pp] != 0)

                        ]

                        # # M-coefficient computation for point p :
                        num = sum([(1.0 / downscaleImg.array[pp]) for pp, _ in pp_tortuosity])
                        denom = sum(
                            [
                                1.0 / (downscaleImg.array[pp] * tortuosity)
                                for pp, tortuosity in pp_tortuosity
                            ]
                        )
                        m_coeff = num / denom if denom != 0 else 0

                        c_n.append(m_coeff)
                        print(f"M-coefficient for point {p} result is {m_coeff}")
                        lines.append(f"M-coefficient for point {p} result is {m_coeff}\n")
                        # re-init seed image for next seed point
                        seed_binary.array[p] = 0

                        # sheet_results.cell(firstRow+nbDistMap, 1).value = str(p)
                        # sheet_results.cell(firstRow+nbDistMap, 2).value = m_coeff
                        table.append([nbDistMap, str(p), m_coeff,""])

                    # M-scalar computation
                    print(f"M-scalar computation:")
                    lines.append(f"M-scalar computation:\n")
                    vrb.mainWindow.groupBoxProcessing.setText(
                        "M-scalar computation ...")
                    qt.QApplication.processEvents()

                    center = self.center_of_mass(in_binary)
                    print(f"Center of Mass is {center}")
                    lines.append(f"Center of Mass is {center}\n")

                    num = sum(
                        [
                            1.0 / self.euclidean_dist(p, center)
                            for idx_p, p in enumerate(valid_points)
                            if c_n[idx_p] != 0
                        ]
                    )
                    denom = sum(
                        [
                            1.0 / (self.euclidean_dist(p, center) * c_n[idx_p])
                            for idx_p, p in enumerate(valid_points)
                            if c_n[idx_p] != 0
                        ],
                    )
                    m_scalar = num / denom if denom != 0 else 0

                    table.append(["","","",""])
                    table.append(["","","",""])
                    table.append(["M-Scalar","Center of Mass X","Center of Mass Y","Center of Mass Z"])
                    table.append([m_scalar,str(center[0]),str(center[1]),str(center[2])])

                    print(f"M-scalar result: {m_scalar}")
                    lines.append(f"M-scalar result: {m_scalar}\n")
                    end = time.time()
                    print(f"Elapsed time is : {(end - start)}")
                    lines.append(f"Elapsed time is : {(end - start)} seconds.\n")

                    table.append(["","","",""])
                    table.append(["Elapsed time",str(end-start),"",""])

                    # SAVE DATAS OF TORTUOSITY IN TABLE
                    output = Measure()
                    output.subType = "Table"
                    output.value = table
                    outputDict = {}
                    outputDict["Name"] = "Table"
                    outputDict["Type"] = "Measure"
                    functionXmlElement = xmlet.Element('FunctionCall')
                    Dfct.SubElement(functionXmlElement, 'Name').text = 'UserMacro'
                    labelValue, realOutputName = vrb.mainWindow.addResult(output, outputDict, functionXmlElement)

                    # sheet_results.cell(firstRow + int(self.lineEditNbSeeds.text()) + 2, 1).value = m_scalar
                    # sheet_results.cell(firstRow + int(self.lineEditNbSeeds.text()) + 2, 2).value = str(center)
                    # sheet_results.cell(firstRow + int(self.lineEditNbSeeds.text()) + 5, 2).value = end - start

                    # Création des bordures du tableau
                    # thin = Side(border_style="thin", color="000000")  # border style, color
                    # border = Border(left=thin, right=thin, top=thin, bottom=thin)  # the position of the border
                    # rows = "A14:B" + str(firstRow + len(valid_points))
                    # for row in sheet_results[rows]:
                    #     for cell in row:
                    #         cell.border = border  # A5:D6 area cells set borders

                    try:
                        if saveExcel:
                            # save the excel file
                            # workbook.save(savePath)
                            with open(savePath, 'w') as f:
                                for line in lines:
                                    f.write(line)
                                    f.write('\n')

                            self.messageBoxEndOfMacro = wgt.MessageBox(
                                "Results saved in file " + savePath, '',
                                buttons=[qt.QMessageBox.Ok], icon=qt.QMessageBox.Information, windowTitle="Info")
                            self.messageBoxEndOfMacro.exec()

                    except:
                        traceback.print_exc(file=sys.stderr)
                        self.messageBox = wgt.MessageBox(
                            "Unable to save Excel file, another instance is already open. Close the excel file before restarting the process.",
                            '', buttons=[qt.QMessageBox.Ok], icon=qt.QMessageBox.Warning, windowTitle="Error")
                        self.messageBox.exec()

            except:
                traceback.print_exc(file=sys.stderr)
                self.messageBox = wgt.MessageBox("An error occured and the process stopped.", '', buttons=[qt.QMessageBox.Ok], icon=qt.QMessageBox.Warning, windowTitle="Error")
                self.messageBox.exec()

        try:
            vrb.mainWindow.groupBoxProcessing.buttonStop.setVisible(False)
        except:
            pass

        try:
            vrb.mainWindow.toggleGroupBoxProcessing()
        except:
            pass

# FILE GROUPBOX TO CHOOSE DIRECTORY
class FileSelectorSave(qt.QGroupBox):

    def __init__(self):
        qt.QGroupBox.__init__(self)

        self.labelFile = qt.QLabel("Backup folder :")
        self.labelFile.setFixedWidth(120 * vrb.ratio)
        self.lineEditFile = qt.QLineEdit()
        self.lineEditFile.setFixedSize(180 * vrb.ratio, 20*vrb.ratio)
        self.buttonOpen = wgt.PushButtonImage(vrb.folderImages + "/Folder.png")
        self.buttonOpen.setFixedSize(20 * vrb.ratio, 20 * vrb.ratio)

        groupBoxFile = qt.QGroupBox()
        self.layoutFile = qt.QGridLayout()
        self.layoutFile.setContentsMargins(10, 10, 10, 10)
        self.layoutFile.addWidget(self.lineEditFile, 0, 0)
        self.layoutFile.addWidget(self.buttonOpen, 0, 1, Qt.AlignRight)
        groupBoxFile.setLayout(self.layoutFile)

        self.layout = qt.QGridLayout()
        self.layout.setContentsMargins(10, 10, 10, 10)
        self.layout.addWidget(self.labelFile, 0, 0)
        self.layout.addWidget(groupBoxFile, 1, 0)
        self.setLayout(self.layout)

        self.layout.setContentsMargins(0, 0, 0, 0)

        style = fct.getStyleSheet()
        self.setStyleSheet(style)


# FILE SELECTOR FOR CSV
class FileSelectorCSV(qt.QGroupBox):

    def __init__(self):
        qt.QGroupBox.__init__(self)

        self.labelFile = qt.QLabel("CSV file :")
        self.labelFile.setFixedWidth(120 * vrb.ratio)
        self.lineEditFile = qt.QLineEdit()
        self.lineEditFile.setFixedSize(180 * vrb.ratio, 20*vrb.ratio)
        self.buttonOpen = wgt.PushButtonImage(vrb.folderImages + "/Folder.png")
        self.buttonOpen.setFixedSize(20 * vrb.ratio, 20 * vrb.ratio)

        groupBoxFile = qt.QGroupBox()
        self.layoutFile = qt.QGridLayout()
        self.layoutFile.setContentsMargins(10, 10, 10, 10)
        self.layoutFile.addWidget(self.lineEditFile, 0, 0)
        self.layoutFile.addWidget(self.buttonOpen, 0, 1, Qt.AlignRight)
        groupBoxFile.setLayout(self.layoutFile)

        self.layout = qt.QGridLayout()
        self.layout.setContentsMargins(10, 10, 10, 10)
        self.layout.addWidget(self.labelFile, 0, 0)
        self.layout.addWidget(groupBoxFile, 1, 0)
        self.setLayout(self.layout)

        self.layout.setContentsMargins(0, 0, 0, 0)

        style = fct.getStyleSheet()
        self.setStyleSheet(style)


mtortuosityWidget = MTortuosityWidget()

def openWidgetMicroporosities():
    mtortuosityWidget.loadModeType()
    fct.showWidget(mtortuosityWidget)

try:
    buttonProcessMTortuosity.clicked.connect(openWidgetMicroporosities)
except:
    pass


if __name__ == '__main__':
    import sys

    app = QtCore.QCoreApplication.instance()
    if app is None:
        app = qt.QApplication([])

    sys._excepthook = sys.excepthook


    def exception_hook(exctype, value, traceback):
        print(exctype, value, traceback)
        sys._excepthook(exctype, value, traceback)
        sys.exit(1)

    sys.excepthook = exception_hook

    # foo = WidgetIDS()
    # foo.show()
    # foo.camera.display_image()
    mtortuosityWidget = MTortuosityWidget()
    mtortuosityWidget.loadModeType()
    foo = mtortuosityWidget
    foo.resize(400,600)
    foo.show()

    app.exec_()